#include "blitz_class.h"

#define REGISTER_GROW			256

static BBClass		**RegisterBase;
static BBClass		**RegisterPut;
static BBClass		**RegisterEnd;

//	The maximum number of registered classes is really not that important,
//	as reaching it would require excessive overuse of classes.
//	Added just to be sure anyway. Used only in debug mode.

#define MAX_NUM_REGISTERED	BBSIZE_MAX

#ifdef BB_DEBUG
static BBSize	NumRegistered_debug;
#endif

void bbClassRegister(const BBClass *clas)
{
	BB_ASSERT(NumRegistered_debug < MAX_NUM_REGISTERED, "Too many registered classes (max = %i)", MAX_NUM_REGISTERED);
	BB_ASSERT(clas->instance_size < BBCLASS_MAX_INSTANCESIZE, "Class instance size too large");
	
	if (RegisterPut == RegisterEnd)
	{
		ptrdiff_t len  = RegisterPut - RegisterBase;
		size_t new_len = len + REGISTER_GROW;
		
		RegisterBase = bbMemExtend(RegisterBase, len * sizeof (BBClass*),
								   new_len * sizeof (BBClass*));
		RegisterEnd = RegisterBase + new_len;
		RegisterPut = RegisterBase + len;
	}
	
	*RegisterPut++ = clas;
#ifdef BB_DEBUG
	++NumRegistered_debug;
#endif
}

BBClass** bbClassAllRegistered(BBInt *count)
{
	*count = RegisterPut - RegisterBase;
	return RegisterBase;
}

void bb_class_register_multiple(unsigned int count, ...)
{
	BB_ASSERT(NumRegistered_debug + count < MAX_NUM_REGISTERED, "Too many registered classes.");
	
	if (RegisterEnd - RegisterPut < count)
	{
		ptrdiff_t len  = RegisterPut - RegisterBase;
		size_t new_len = len + REGISTER_GROW;
		
		RegisterBase = bbMemExtend(RegisterBase, len * sizeof (BBClass*),
								   new_len * sizeof (BBClass*));
		RegisterEnd = RegisterBase + new_len;
		RegisterPut = RegisterBase + len;
	}	
	
	va_list args;
	
	const BBClass *clas;
	
	va_start(args, count);
	
	for (uint_fast32_t c = 0; c < count; ++c)
	{
		clas = va_arg(args, BBClass*);
		
		BB_ASSERT(clas->instance_size < BBCLASS_MAX_INSTANCESIZE, "Class \"%s\" instance size too large", clas->debug_scope->name);
		*RegisterPut++ = clas;
	}

	va_end(args);
}
